#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/reboot.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <linux/io.h>
#include <linux/module.h>
#include <asm/setup.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/kthread.h>

struct crash_log_dump {
	unsigned int size;
	unsigned char *buffer;
};

static struct crash_log_dump crash_dump_log;

static unsigned int kmsg_size;
static unsigned char *kmsg_buffer;
static unsigned int main_size;
static unsigned char *main_buffer;
static unsigned int event_size;
static unsigned char *event_buffer;
static unsigned int sys_size;
static unsigned char *sys_buffer;
extern unsigned long log_main_buffer_address;
extern unsigned long log_events_buffer_address;
extern unsigned long log_system_buffer_address; 
static unsigned char local_main_buf[1024]; 
static unsigned int crash_buf_size;
static int crash_store_flag;
static int panic_flag = 0;
extern panic_reset(void);

void set_crash_store_enable(void)
{
	crash_store_flag = 1;
}

void set_crash_store_disable(void)
{
	crash_store_flag = 0;
}

void store_crash_log(const char *logname)
{
	unsigned char *p = NULL;
	long i;

	if (!crash_store_flag)
		return;

	crash_dump_log.size = 0;

	if(!strcmp(logname,"Mpanic_data")){
		 crash_dump_log.buffer = main_buffer;
		 p = __va(log_main_buffer_address);
        	*p = 0xDA; 
	}
	else if(!strncmp(logname,"Epanic_data",5)){
		 crash_dump_log.buffer = event_buffer;
		 p = __va(log_events_buffer_address);
	}
	else if(!strncmp(logname,"Spanic_data",5)){
		 crash_dump_log.buffer = sys_buffer;
		 p = __va(log_system_buffer_address);
		crash_store_flag = 0;
	}

 
	memcpy(crash_dump_log.buffer, p, (SZ_256K-2));
	if (!crash_store_flag){
		panic_flag = 1;
	}

       	return;
}

static ssize_t sys_android_read_old(struct file *file, char __user *buf,
		size_t len, loff_t *offset)
{
	loff_t pos = *offset;
	ssize_t count;

	if (pos >= SZ_256K)
		return 0;

	count = min(len, (size_t)(SZ_256K - pos));
	if (copy_to_user(buf, sys_buffer+ pos, count))
		return -EFAULT;

	*offset += count;
	return count;
}

static ssize_t event_android_read_old(struct file *file, char __user *buf,
		size_t len, loff_t *offset)
{
	loff_t pos = *offset;
	ssize_t count;

	if (pos >= SZ_256K)
		return 0;

	count = min(len, (size_t)(SZ_256K - pos));
	if (copy_to_user(buf, event_buffer+ pos, count))
		return -EFAULT;

	*offset += count;
	return count;
}

static ssize_t main_android_reset_crash_bit(struct file *file, const char __user *buf,
                size_t len, loff_t *offset)
{
        printk("reset crash bit\n"); 
       	*main_buffer = 0x00;
        return 1;
}

static ssize_t main_android_read_old(struct file *file, char __user *buf,
		size_t len, loff_t *offset)
{
	loff_t pos = *offset;
	ssize_t count, i;

	if (pos >= SZ_1K)
		return 0;

	count = min(len, (size_t)(SZ_256K - pos));
	if (copy_to_user(buf, main_buffer + pos, count))
		return -EFAULT;

	*offset += count;
	return count;
}

static int restore_crash_log(struct notifier_block *this, unsigned long event,
		void *ptr)
{
	unsigned long flags;
	printk("%s\n",__func__);
	panic_reset(); //reset the wdt

	return NOTIFY_DONE;
}

static struct notifier_block panic_handler_block = {
	.notifier_call  = restore_crash_log,
};

static const struct file_operations sys_android_file_ops = {
	.owner = THIS_MODULE,
	.read = sys_android_read_old,
};

static const struct file_operations event_android_file_ops = {
	.owner = THIS_MODULE,
	.read = event_android_read_old,
};

static const struct file_operations main_android_file_ops = {
	.owner = THIS_MODULE,
	.read = main_android_read_old,
	.write = main_android_reset_crash_bit,
};

static int panic_thread(void *data){
	int i;
	char *y = NULL;
	printk("%s\n",__func__);


	mdelay(120000);
	printk("%s:Generating Panic \n",__func__);
	printk("panic_thread: %c\n",(*y));
	while(panic_flag == 0);

	return 0;
}

static int __init panic_handler_probe(struct platform_device *pdev)
{
	struct resource *res = pdev->resource;
	struct proc_dir_entry *m_entry;
	struct proc_dir_entry *e_entry;
	struct proc_dir_entry *s_entry;
	size_t start;
	size_t buffer_size;
	int ret = 0, i;
	char *p = __va(log_main_buffer_address);
	struct task_struct *th;


	if (res == NULL || pdev->num_resources != 3 ||
			!(res->flags & IORESOURCE_MEM)) {
		printk(KERN_ERR "panic_handler: invalid resource, %p %d flags "
       				"%lx\n", res, pdev->num_resources, res ? res->flags : 0);
		return -ENXIO;
	}

       	main_size = res[0].end - res[0].start + 1;
	start = res[0].start;
	printk(KERN_INFO "panic_handler: got main_buffer at %zx, size %zx\n",
       			start, main_size);
	main_buffer = ioremap(res[0].start, main_size);
	if (main_buffer == NULL) {
		printk(KERN_ERR "panic_handler: failed to map memory\n");
		return -ENOMEM;
	}

       	event_size = res[1].end - res[1].start + 1;
	start = res[1].start;
	printk(KERN_INFO "panic_handler: got event_buffer at %zx, size %zx\n",
       			start, event_size);
	event_buffer = ioremap(res[1].start, event_size);
	if (event_buffer == NULL) {
		printk(KERN_ERR "panic_handler: failed to map memory\n");
		return -ENOMEM;
	}

       	sys_size = res[2].end - res[2].start + 1;
	start = res[2].start;
	printk(KERN_INFO "panic_handler: got sys_buffer at %zx, size %zx\n",
       			start, sys_size);
	sys_buffer = ioremap(res[2].start, sys_size);
	if (sys_buffer == NULL) {
		printk(KERN_ERR "panic_handler: failed to map memory\n");
		return -ENOMEM;
	}

	m_entry = create_proc_entry("last_main_msg", S_IFREG | S_IRUGO, NULL);
	if (!m_entry) {
		printk(KERN_ERR "last_main_msg: failed to create proc entry\n");
		return 0;
	}

	m_entry->proc_fops = &main_android_file_ops;
	m_entry->size = SZ_256K;

	e_entry = create_proc_entry("last_event_msg", S_IFREG | S_IRUGO, NULL);
	if (!e_entry) {
		printk(KERN_ERR "last_event_msg: failed to create proc entry\n");
		return 0;
	}

	e_entry->proc_fops = &event_android_file_ops;
	e_entry->size = SZ_256K;

	s_entry = create_proc_entry("last_sys_msg", S_IFREG | S_IRUGO, NULL);
	if (!s_entry) {
		printk(KERN_ERR "last_sys_msg: failed to create proc entry\n");
		return 0;
	}

	s_entry->proc_fops = &sys_android_file_ops;
	s_entry->size = SZ_256K;

	/* Setup panic notifier */
	atomic_notifier_chain_register(&panic_notifier_list, &panic_handler_block);

/*	th = kthread_run(panic_thread, NULL, "panic-thread");
	if (IS_ERR(th))
	{
		pr_info("Unable to start control thread\n");
		return PTR_ERR(th);
	}
*/
	return ret;
}	

static int __devexit panic_handler_remove(struct platform_device *pdev)
{
	return 0;
}

static struct platform_driver panic_handler_driver __refdata = {
	.probe = panic_handler_probe,
	.remove = __devexit_p(panic_handler_remove),
	.driver = {
		.name = "panic-handler",
		.owner = THIS_MODULE,
	},
};

static int __init panic_handler_init(void)
{
	return platform_driver_register(&panic_handler_driver);
}

static void __exit panic_handler_exit(void)
{
	platform_driver_unregister(&panic_handler_driver);
}

module_init(panic_handler_init);
module_exit(panic_handler_exit);
