/*
 * MAXWELL DVF9919 Soc Codec Driver
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/clk.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/fsl_devices.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/asoundef.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <mach/hardware.h>

static int mxc_ar3002_codec_startup(struct snd_pcm_substream *substream,
				  struct snd_soc_dai *dai)
{
  printk(KERN_INFO"%s(%i)\n", __FUNCTION__, __LINE__);
	return 0;
}

static int mxc_ar3002_codec_prepare(struct snd_pcm_substream *substream,
				  struct snd_soc_dai *dai)
{
  printk(KERN_INFO"%s(%i)\n", __FUNCTION__, __LINE__);
	return 0;
}

static void mxc_ar3002_codec_shutdown(struct snd_pcm_substream *substream,
				   struct snd_soc_dai *dai)
{
  printk(KERN_INFO"%s(%i)\n", __FUNCTION__, __LINE__);
}

static struct snd_soc_dai_ops mxc_ar3002_codec_dai_ops = {
	.startup = mxc_ar3002_codec_startup,
	.prepare = mxc_ar3002_codec_prepare,
	.shutdown = mxc_ar3002_codec_shutdown,
};

#define AR3002_RATES (SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000)

#define AR3002_FORMATS (SNDRV_PCM_FMTBIT_MU_LAW | SNDRV_PCM_FMTBIT_A_LAW | \
			SNDRV_PCM_FMTBIT_S8 | SNDRV_PCM_FMTBIT_U8 | SNDRV_PCM_FMTBIT_S16_LE)
	

static struct snd_soc_dai_driver mxc_ar3002_codec_dai = {
	.name = "ar3002",
	.playback = {
                .stream_name = "Playback",
                .channels_min = 1,
                .channels_max = 1,
                .rates = AR3002_RATES,
                .formats = AR3002_FORMATS,
        },
        .capture = {
                .stream_name = "Capture",
                .channels_min = 1,
                .channels_max = 1,
                .rates = AR3002_RATES,
                .formats = AR3002_FORMATS,
        },
	.ops = &mxc_ar3002_codec_dai_ops,
};

static int mxc_ar3002_codec_soc_probe(struct snd_soc_codec *codec)
{
	return 0;
}

static struct snd_soc_codec_driver soc_codec_dev_ar3002 = {
	.probe = mxc_ar3002_codec_soc_probe,
};

static int __devinit mxc_ar3002_codec_probe(struct platform_device *pdev)
{
	int ret = 0;
printk(KERN_INFO"%s(%i)\n", __FUNCTION__, __LINE__);
	dev_info(&pdev->dev, "MXC AR3002 Audio\n");
printk(KERN_INFO"%s(%i)\n", __FUNCTION__, __LINE__);
	ret = snd_soc_register_codec(&pdev->dev, &soc_codec_dev_ar3002,
				     &mxc_ar3002_codec_dai, 1);
printk(KERN_INFO"%s(%i)\n", __FUNCTION__, __LINE__);
	if (ret) {
		dev_err(&pdev->dev, "failed to register codec\n");
		return ret;
	}

	return 0;
}

static int __devexit mxc_ar3002_codec_remove(struct platform_device *pdev)
{
	snd_soc_unregister_codec(&pdev->dev);
	platform_set_drvdata(pdev, NULL);

	return 0;
}

struct platform_driver mxc_ar3002_driver = {
	.driver = {
		   .name = "ar3002",
		   .owner = THIS_MODULE,
		   },
	.probe = mxc_ar3002_codec_probe,
	.remove = __devexit_p(mxc_ar3002_codec_remove),
};

static int __init mxc_ar3002_codec_init(void)
{
    printk(KERN_INFO"%s(%i)\n", __FUNCTION__, __LINE__);
	return platform_driver_register(&mxc_ar3002_driver);
}

static void __exit mxc_ar3002_codec_exit(void)
{
	return platform_driver_unregister(&mxc_ar3002_driver);
}

module_init(mxc_ar3002_codec_init);
module_exit(mxc_ar3002_codec_exit);

MODULE_AUTHOR("Gigaset Communications GmbH");
MODULE_DESCRIPTION("Maxwell AR3002 Audio");
MODULE_LICENSE("GPL");
