/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/bitops.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/err.h>
#include <linux/irq.h>
#include <linux/io.h>
#include <linux/fsl_devices.h>
#include <linux/slab.h>
#include <linux/clk.h>
#include <linux/switch.h>
#include <linux/kthread.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <sound/jack.h>
#include <mach/dma.h>
#include <mach/clock.h>
#include <mach/audmux.h>
#include <mach/gpio.h>
#include <asm/mach-types.h>

#include "imx-ssi.h"

struct ar3002_imx_priv {
	int sysclk;         /*mclk from the outside*/
	int codec_sysclk;
	int dai_hifi;
	struct platform_device *pdev;
	struct snd_pcm_substream *first_stream;
	struct snd_pcm_substream *second_stream;
};
static unsigned int sample_format = SNDRV_PCM_FMTBIT_S16_LE;
static struct ar3002_imx_priv card_priv;
static struct snd_soc_card snd_soc_card_imx;
static struct snd_soc_codec *gcodec;

static int imx_ar3002_startup(struct snd_pcm_substream *substream)
{
	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);
	return 0;
}

static void imx_ar3002_shutdown(struct snd_pcm_substream *substream)
{
	return;
}

static int imx_ar3002_hw_params(struct snd_pcm_substream *substream,
				     struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct ar3002_imx_priv *priv = &card_priv;
	unsigned int channels = params_channels(params);
	int ret = 0;
	u32 dai_format;
	unsigned int pll_out;

	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);
	if (!priv->first_stream)
		priv->first_stream = substream;
	else
		priv->second_stream = substream;

	/* set i.MX active slot mask */
	// snd_soc_dai_set_tdm_slot(cpu_dai,0xfffffffc, 0xfffffffc, 2, 16);

	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);

	ret = snd_soc_dai_set_sysclk(cpu_dai, 0, 0, SND_SOC_CLOCK_IN);
        if (ret < 0) {
                pr_err("Failed to set SYSCLK: %d\n", ret);
                return ret;
        }
	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);

	dai_format = SND_SOC_DAIFMT_I2S | SND_SOC_DAIFMT_NB_NF |
		SND_SOC_DAIFMT_CBM_CFM;

	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);
	/* set cpu DAI configuration */
	ret = snd_soc_dai_set_fmt(cpu_dai, dai_format);
	if (ret < 0)
		return ret;

	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);
	return 0;
}

static int imx_ar3002_hw_free(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->codec_dai;
	struct ar3002_imx_priv *priv = &card_priv;
	int ret;

	if (priv->first_stream == substream)
		priv->first_stream = priv->second_stream;
	priv->second_stream = NULL;

	return 0;
}

static void imx_resume_event(struct work_struct *wor)
{
	return;
}

static DECLARE_DELAYED_WORK(resume_hp_event, imx_resume_event);

int imx_ar3002_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct ar3002_imx_priv *priv = &card_priv;
	struct platform_device *pdev = priv->pdev;
	struct mxc_audio_platform_data *plat = pdev->dev.platform_data;

	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);
	if (SNDRV_PCM_TRIGGER_RESUME == cmd) {
			schedule_delayed_work(&resume_hp_event,
				msecs_to_jiffies(200));
	}

	return 0;
}

static int imx_ar3002_init(struct snd_soc_pcm_runtime *rtd)
{
	printk(KERN_INFO"Maxwell--%s(%i)\n",__FUNCTION__, __LINE__);
	return 0;
}

static struct snd_soc_ops imx_hifi_ops = {
	.startup = imx_ar3002_startup,
	.shutdown = imx_ar3002_shutdown,
	.hw_params = imx_ar3002_hw_params,
	.hw_free = imx_ar3002_hw_free,
	.trigger = imx_ar3002_trigger,
};

static struct snd_soc_dai_link imx_dai[] = {
	{
		.name = "AR3002",
		.stream_name = "AR3002",
		.codec_dai_name = "ar3002",
                .codec_name     = "ar3002.0",
		.cpu_dai_name	= "imx-ssi.2",
		.platform_name	= "imx-pcm-audio.2",
		.init		= imx_ar3002_init,
		.ops		= &imx_hifi_ops,
	},
};

static struct snd_soc_card snd_soc_card_imx = {
	.name		= "ar3002-audio",
	.dai_link	= imx_dai,
	.num_links	= ARRAY_SIZE(imx_dai),
};

static int imx_audmux_config(int slave, int master)
{
	unsigned int ptcr, pdcr;
	slave = slave - 1;
	master = master - 1;

	printk(KERN_INFO"Maxwell--%s_ar3002(%i)  slave=%i   master=%i\n",__FUNCTION__, __LINE__, slave, master);
	
	
	ptcr = MXC_AUDMUX_V2_PTCR_SYN |
		MXC_AUDMUX_V2_PTCR_TFSDIR |
		MXC_AUDMUX_V2_PTCR_TFSEL(master) |
		MXC_AUDMUX_V2_PTCR_TCLKDIR |
		MXC_AUDMUX_V2_PTCR_TCSEL(master);
	pdcr = MXC_AUDMUX_V2_PDCR_RXDSEL(master);
	mxc_audmux_v2_configure_port(slave, ptcr, pdcr);

	ptcr = MXC_AUDMUX_V2_PTCR_SYN;
	pdcr = MXC_AUDMUX_V2_PDCR_RXDSEL(slave);
	mxc_audmux_v2_configure_port(master, ptcr, pdcr);

	return 0;
}

/*
 * This function will register the snd_soc_pcm_link drivers.
 */
static int __devinit imx_ar3002_probe(struct platform_device *pdev)
{

	struct mxc_audio_platform_data *plat = pdev->dev.platform_data;
	struct ar3002_imx_priv *priv = &card_priv;
	int ret = 0;

	priv->pdev = pdev;

	imx_audmux_config(plat->src_port, plat->ext_port);

	if (plat->init && plat->init()) {
		ret = -EINVAL;
		return ret;
	}

	priv->sysclk = plat->sysclk;

	priv->first_stream = NULL;
	priv->second_stream = NULL;
	return ret;
}

static int __devexit imx_ar3002_remove(struct platform_device *pdev)
{
	struct mxc_audio_platform_data *plat = pdev->dev.platform_data;
	struct ar3002_imx_priv *priv = &card_priv;

	plat->clock_enable(0);

	if (plat->finit)
		plat->finit();

	return 0;
}

static struct platform_driver imx_ar3002_driver = {
	.probe = imx_ar3002_probe,
	.remove = imx_ar3002_remove,
	.driver = {
		   .name = "maxwell-ar3002",
		   .owner = THIS_MODULE,
		   },
};

static struct platform_device *imx_snd_device;

static int __init imx_ar3002_asoc_init(void)
{
	int ret;

	ret = platform_driver_register(&imx_ar3002_driver);
	if (ret < 0)
		goto exit;

	imx_snd_device = platform_device_alloc("soc-audio", 2);
	if (!imx_snd_device)
		goto err_device_alloc;

	platform_set_drvdata(imx_snd_device, &snd_soc_card_imx);

	ret = platform_device_add(imx_snd_device);

	if (0 == ret)
		goto exit;

	platform_device_put(imx_snd_device);

err_device_alloc:
	platform_driver_unregister(&imx_ar3002_driver);
exit:
	return ret;
}

static void __exit imx_ar3002_asoc_exit(void)
{
	platform_driver_unregister(&imx_ar3002_driver);
	platform_device_unregister(imx_snd_device);
}

module_init(imx_ar3002_asoc_init);
module_exit(imx_ar3002_asoc_exit);

/* Module information */
MODULE_AUTHOR("Gigaset Communications GmbH");
MODULE_DESCRIPTION("ALSA SoC maxwell ar3002");
MODULE_LICENSE("GPL");
